package Sub::Uplevel;

use 5.006;

use strict;
use vars qw($VERSION @ISA @EXPORT);
$VERSION = 0.02;

# We have to do this so the CORE::GLOBAL versions override the builtins
_restore_CORE_GLOBAL();

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(uplevel);

=head1 NAME

Sub::Uplevel - run a function in a higher stack frame

=head1 SYNOPSIS

  use Sub::Uplevel;

  sub foo {
      print join " - ", caller;
  }

  sub bar {
      uplevel 1, \&foo;
  }

  #line 11
  bar();    # main - foo.plx - 11

=head1 DESCRIPTION

Like Tcl's uplevel() function, but not quite so dangerous.


=over 4

=item B<uplevel>

  uplevel $num_frames, \&func, @args;

Makes the given function think it's being executed $num_frames higher
than the current stack level.  So caller() will be caller($num_frames)
for them.

=cut

sub uplevel {
    my($num_frames, $func, @args) = @_;
    $num_frames += 2;

    no warnings 'redefine';

    my $old_caller = \&CORE::GLOBAL::caller;
    local *CORE::GLOBAL::caller = sub {
        my $height = $_[0] || 0;
        $height += $num_frames;
        return undef if $height <= 0;
        my @caller = $old_caller->($height + 1);

        if( wantarray ) {
            if( !@_ ) {
                @caller = @caller[0..2];
            }
            return @caller;
        }
        else {
            return $caller[0];
        }
    };

    my $old_die = \&CORE::GLOBAL::die;
    local *CORE::GLOBAL::die = sub {
        if( @_ and $_[0] !~ /\n$/ ) {
            $old_die->(sprintf("%s at %s line %d.\n", join('', @_), 
                               (caller($num_frames))[1,2]));
        }
        else {
            CORE::die @_;
        }
    };

    my $old_warn = \&CORE::GLOBAL::warn;
    local *CORE::GLOBAL::warn = sub {
        if( @_ and $_[0] !~ /\n$/ ) {
            $old_warn->(sprintf("%s at %s line %d.\n", join('', @_), 
                                (caller($num_frames))[1,2]));
        }
        else {
            CORE::warn @_;
        }
    };


    my @caller = CORE::caller($num_frames-2);
    my $wantarray = $caller[5];

    my($output, @output);
    if( !defined $wantarray ) {
        $func->(@args);
    }
    elsif( !$wantarray ) {
        $output = $func->(@args);
    }
    else {
        @output = $func->(@args);
    }

    return $wantarray ? @output : $output;
}


sub _restore_CORE_GLOBAL {
    no warnings 'redefine';

    *CORE::GLOBAL::caller = sub {
        my $height = $_[0] || 0;
        return undef if $height < 0;
        my @caller = CORE::caller($height);

        if( wantarray ) {
            if( !@_ ) {
                @caller = @caller[0..2];
            }
            return @caller;
        }
        else {
            return $caller[0];
        }
    } unless *CORE::GLOBAL::caller{CODE};

    my $old_die = \&CORE::GLOBAL::die;
    *CORE::GLOBAL::die = sub {
        if( @_ and $_[0] !~ /\n$/ ) {
            $old_die->(sprintf("%s at %s line %d.\n", join('', @_), 
                               (caller(1))[1,2]));
        }
        else {
            CORE::die @_;
        }
    };

    my $old_warn = \&CORE::GLOBAL::warn;
    *CORE::GLOBAL::warn = sub {
        if( @_ and $_[0] !~ /\n$/ ) {
            $old_warn->(sprintf("%s at %s line %d.\n", join('', @_), 
                                (caller(1))[1,2]));
        }
        else {
            CORE::warn @_;
        }
    };
}


=back

=head1 BUGS and CAVEATS

Symbol::Uplevel must be used before any code which uses it.

=cut


1;
